from __future__ import annotations
import asyncio
import polars as pl
from typing import Iterable
from navpipe.transport.async_http import AsyncNavPipeHTTP
from navpipe.transform.nav import nav_response_to_df
from navpipe.models.nav import NavResponse, DateRange
import msgspec


async def _fetch_nav_df(
    http: AsyncNavPipeHTTP,
    scheme_code: int,
    date_range: DateRange | None,
    sem: asyncio.Semaphore,
) -> pl.DataFrame:
    params: dict[str, str] = {}
    if date_range:
        params["startDate"] = date_range.start_date
        params["endDate"] = date_range.end_date

    async with sem:
        raw_bytes = await http.get_raw(f"/mf/{scheme_code}", params=params)

    # Decode directly into NavResponse
    typed = msgspec.json.decode(raw_bytes, type=NavResponse)

    return nav_response_to_df(typed)


async def fetch_nav_history_bulk(
    scheme_codes: Iterable[int],
    *,
    date_range: DateRange | None = None,
    max_concurrency: int = 5,
    rate_limit_per_sec: int | None = None,
) -> pl.DataFrame:
    sem = asyncio.Semaphore(max_concurrency)

    async with AsyncNavPipeHTTP(rate_limit_per_sec=rate_limit_per_sec) as http:
        tasks = [_fetch_nav_df(http, code, date_range, sem) for code in scheme_codes]

        dfs = await asyncio.gather(*tasks)

    return pl.concat(dfs, how="vertical")



from __future__ import annotations
import aiohttp
import asyncio
import random
import msgspec
from aiolimiter import AsyncLimiter
import types


class AsyncNavPipeHTTP:
    BASE_URL = "https://api.mfapi.in"

    def __init__(
        self,
        *,
        timeout: int = 20,
        rate_limit_per_sec: int | None = None,
        max_retries: int = 4,
    ):
        self.timeout: aiohttp.ClientTimeout = aiohttp.ClientTimeout(total=timeout)
        self._session: "aiohttp.ClientSession | None" = None
        self._limiter: "AsyncLimiter | None" = (
            AsyncLimiter(rate_limit_per_sec, time_period=1)
            if rate_limit_per_sec
            else None
        )
        self._max_retries: int = max_retries
        self._decoder = msgspec.json.Decoder()

    async def __aenter__(self) -> AsyncNavPipeHTTP:
        self._session = aiohttp.ClientSession(timeout=self.timeout)
        return self

    async def __aexit__(
        self,
        _exc_type: type | None,
        _exc_value: BaseException | None,
        _tb: "types.TracebackType | None",
    ) -> None:
        _ = (_exc_type, _exc_value, _tb)
        if self._session:
            await self._session.close()
            self._session = None

    async def get_raw(
        self, endpoint: str, params: dict[str, str] | None = None
    ) -> bytes:
        if self._limiter:
            async with self._limiter:
                return await self._request_raw(endpoint, params)
        return await self._request_raw(endpoint, params)

    async def get(
        self, endpoint: str, params: dict[str, str] | None = None
    ) -> dict[str, object]:
        raw_bytes = await self.get_raw(endpoint, params)
        return self._decoder.decode(raw_bytes)

    async def _request_raw(
        self, endpoint: str, params: dict[str, str] | None = None
    ) -> bytes:
        if not self._session:
            raise RuntimeError(
                "HTTP session not initialized. Use async context manager."
            )

        url = f"{self.BASE_URL}{endpoint}"

        for attempt in range(self._max_retries):
            try:
                async with self._session.get(url, params=params) as resp:
                    if resp.status in {429, 500, 502, 503, 504}:
                        retry_after = resp.headers.get("Retry-After")
                        if retry_after:
                            await asyncio.sleep(float(retry_after))
                        else:
                            backoff = 2**attempt + random.uniform(0, 0.3)
                            await asyncio.sleep(backoff)
                        continue

                    resp.raise_for_status()
                    return await resp.read()

            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                if attempt < self._max_retries - 1:
                    backoff = 2**attempt + random.uniform(0, 0.3)
                    await asyncio.sleep(backoff)
                    continue
                raise RuntimeError(f"Max retries exceeded for {url}") from e

        raise RuntimeError(f"Max retries exceeded for {url}")



from __future__ import annotations
import msgspec


class NavData(msgspec.Struct):
    date: str
    nav: str


class NavMeta(msgspec.Struct):
    fund_house: str
    scheme_type: str
    scheme_category: str
    scheme_code: int
    scheme_name: str
    isin_growth: str | None
    isin_div_reinvestment: str | None


class NavResponse(msgspec.Struct):
    meta: NavMeta
    data: list[NavData]


class DateRange(msgspec.Struct, frozen=True):
    start_date: str
    end_date: str



from .api import NavPipe
import importlib.metadata

try:
    __version__ = importlib.metadata.version("navpipe")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.1.1"


__all__ = ["NavPipe"]



import polars as pl
from typing import Iterable
from navpipe.dataset.nav import fetch_nav_history_bulk, DateRange
from navpipe.utils.sync import run_async

class NavPipe:
    def __init__(
        self,
        *,
        max_concurrency: int = 5,
        rate_limit_per_sec: int | None = 3,
    ):
        self._max_concurrency = max_concurrency
        self._rate_limit = rate_limit_per_sec

    def nav_history(
        self,
        scheme_codes: Iterable[int],
        *,
        start_date: str | None = None,
        end_date: str | None = None
    ) -> pl.DataFrame:
        
        if (start_date and not end_date) or (end_date and not start_date):
            raise ValueError("Both start_date and end_date must be provided together.")
        
        date_range = None
        if start_date and end_date:
            date_range = DateRange(start_date=start_date, end_date=end_date)

        return run_async(
            fetch_nav_history_bulk(
                scheme_codes=scheme_codes,
                date_range=date_range,
                max_concurrency=self._max_concurrency,
                rate_limit_per_sec=self._rate_limit,
            )
        )



from __future__ import annotations
import polars as pl
from navpipe.models.nav import NavResponse


def nav_response_to_df(resp: NavResponse) -> pl.DataFrame:
    df = pl.DataFrame(
        [
            {
                "scheme_code": int(resp.meta.scheme_code),
                "scheme_name": resp.meta.scheme_name,
                "date": item.date,
                "nav": float(item.nav),
            }
            for item in resp.data
        ]
    )

    return df.with_columns(
        [
            pl.col("date").str.strptime(pl.Date),
        ]
    ).select(["scheme_code", "scheme_name", "date", "nav"])



from __future__ import annotations
import asyncio
import threading
from collections.abc import Coroutine
from typing import TypeVar

T = TypeVar("T")


def run_async(coro: Coroutine[object, object, T]) -> T:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop is None:
        return asyncio.run(coro)

    result: T | None = None

    def runner() -> None:
        nonlocal result
        result = asyncio.run(coro)

    t = threading.Thread(target=runner)
    t.start()
    t.join()

    assert result is not None
    return result
